DatasetQuery = """

top_dataset_query       :    "datasets" dataset_query_list

dataset_query_list: dataset_query ("," dataset_query)*            -> dataset_query_list

?dataset_query   :    dataset_query_with_subsets
    | dataset_query_with_subsets "having" meta_exp          -> dataset_add_where
    
?dataset_query_with_subsets : dataset_spec
    | dataset_spec dataset_provenance_op                    -> dataset_add_subsets

!dataset_provenance_op: "with" "subsets" "recursively"?

!dataset_spec:  qualified_name
    | "matching" "regexp" regexp_pattern
    | "matching" sql_pattern

sql_pattern:    (FNAME ":")? UNQUOTED_STRING                   
regexp_pattern:    (FNAME ":")? STRING                              

?simple_dataset_query : dataset_query_with_subsets

// dataset attributes
DATASET_ATTR_NAME: "namespace" | "name" | "parent_namespace" | "parent_name" | "creator" | "description" | "created_timestamp" | "updated_timestamp"
    | "frozen" 
    | "monotonic"
"""





import sys, getopt, os, json, fnmatch, pprint
from datetime import datetime, timezone
from urllib.parse import quote_plus, unquote_plus
from metacat.util import to_bytes, to_str
from metacat.webapi import MetaCatClient, MCError
from metacat.ui.cli import CLI, CLICommand, InvalidOptions, InvalidArguments

class ListCommand(CLICommand):

    GNUStyle = True
    Opts = "j json"
    Usage = """[options] [<root category>]
        -j|--json           - print as JSON
    """

    def __call__(self, command, client, opts, args):
        root = None if not args else args[0]
        as_json = "-j" in opts or "--json" in opts
        lst = client.list_categories(root)
        if as_json:
            print(json.dumps(lst, indent=4, sort_keys=True))
        else:
            for c in lst:
                print(c["path"])
                
class ShowCommand(CLICommand):
    
    GNUStyle = True
    Opts = "j json"
    MinArgs = 1
    Usage = """[-j|--json] <category>
        -j|--json           - print as JSON
    """
    
    def __call__(self, command, client, opts, args):
        data = client.get_category(args[0])
        if data:
            if "-j" in opts or "--json" in opts:
                print(json.dumps(data, indent=4, sort_keys=True))
            else:
                print("Path:            ", data["path"])
                print("Description:     ", data.get("description", ""))
                print("Owner user:      ", data.get("owner_user", "") or "")
                print("Owner role:      ", data.get("owner_role", "") or "")
                print("Creator:         ", data.get("creator", ""))
                ct = data.get("created_timestamp") or ""
                if ct:
                    ct = datetime.fromtimestamp(ct, timezone.utc).strftime("%Y-%m-%d %H:%M:%S %Z")
                print("Created at:      ", ct)
                print("Restricted:      ", "yes" if data.get("restricted", False) else "no")
                print("Constraints:")
                for name, constraint in sorted(data.get("definitions", {}).items()):
                    line = "  %-40s %10s" % (name, constraint.get("type", "any"))
                    if "values" in constraint:
                        line += " %s" % (tuple(constraint["values"]),)
                    rng = None
                    if "min" in constraint:
                        rng = [repr(constraint["min"]), ""]
                    if "max" in constraint:
                        if rng is None: rng = ["", ""]
                        rng[1] = repr(constraint["max"])
                    if rng is not None:
                        line += " [%s - %s]" % tuple(rng)
                    if "pattern" in constraint:
                        line += " ~ '%s'" % (constraint["pattern"])
                    print(line)



CategoryCLI = CLI(
    "list",     ListCommand(),
    "show",     ShowCommand()
)
 



import json, time, pprint, traceback
from metacat.db import DBDataset, DBFile, DBNamedQuery, DBFileSet
from metacat.util import limited, unique
from metacat.common.trees import Node, Ascender, Descender, Converter
from metacat.common import MetaExpressionDNF
from .sql_converter import SQLConverter
from .query_executor import FileQueryExecutor
from .meta_evaluator import MetaEvaluator
from datetime import date, datetime, timezone

from lark import Lark, LarkError
from lark import Tree, Token

CMP_OPS = [">" , "<" , ">=" , "<=" , "==" , "=" , "!=", "~~", "~~*", "!~~", "!~~*"]

from .grammar import MQL_Grammar
_Parser = Lark(MQL_Grammar, start="query")

class MQLError(Exception):

    def __init__(self, message):
        self.Message = message

class MQLSyntaxError(MQLError):
    
    def __str__(self):
        return f"MQL syntax Error: {self.Message}"
        
class MQLCompilationError(MQLError):
    
    def __str__(self):
        return f"MQL compilation Error: {self.Message}"
        
class MQLExecutionError(MQLError):
    
    def __str__(self):
        return f"MQL execution Error: {self.Message}"
        
def _merge_skip_limit(existing_skip, existing_limit, skip=0, limit=None):
    if existing_limit is None:
        return existing_skip+skip, limit
    elif limit is None:
        return existing_skip + skip, max(0, existing_limit - skip)
    else:
        return existing_skip + skip, max(0, min(existing_limit - skip, limit))
        
class _DatasetQueryCompiler(Ascender):

    def __init__(self):
        Ascender.__init__(self)

    def __call__(self, tree):
        out = self.walk(tree)
        return out

    def dataset_query_list(self, node, *args):
        queries = [a["query"] for a in args]
        sql = DBDataset.sql_for_bdqs(queries)
        return Node("sql", sql=sql)

class _DatasetQueryExecutor(Ascender):
    
    def __init__(self, db):
        Ascender.__init__(self)
        self.DB = db

    def __call__(self, tree):
        out = self.walk(tree)
        return out

    def sql(self, node, sql=None):
        return DBDataset.datasets_from_sql(self.DB, sql)

class DatasetQuery(object):
    
    Type = "dataset"
    
    def __init__(self, tree):
        self.Tree = tree
        self.Compiled = None

    def compile(self, with_meta=False, with_provenance=False):
        self.Compiled = self.Compiled or _DatasetQueryCompiler()(self.Tree)
        return self.Compiled

    def run(self, db, debug=False, **ignore):
        compiled = self.compile()
        return _DatasetQueryExecutor(db)(compiled)


class FileQuery(object):

    Type = "file"

    def __init__(self, tree, include_retired=False):
        self.Tree = tree
        self.Assembled = self.Optimized = self.Compiled = None
        self.IncludeRetired = include_retired

    def __str__(self):
        return "FileQuery(\n%s\n)" % (self.Tree.pretty("  "),)

    def skip_assembly(self):
        if self.Assembled is None:
            self.Assembled = self.Tree
        return self.Assembled

    def optimize(self, debug=False, skip=0, limit=None):
        if self.Optimized is None:
            #print("Query.optimize: assembled:----\n", self.Assembled.pretty())

            optimized = self.Tree
            if debug:
                print("Query.optimize: initial:----")
                print(optimized.pretty("    "))

            #optimized = _SkipLimitApplier().walk(optimized)
            #if debug:
            #    print("Query.optimize: after 1st _SkipLimitApplier:----")
            #    print(optimized.pretty("    "))
                
            #print("starting _MetaExpPusher...")
            optimized = _MetaExpPusher().walk(optimized, None)
            if debug:
                print("Query.optimize: after _MetaExpPusher:----")
                print(optimized.pretty("    "))

            optimized = _RemoveEmpty().walk(optimized, debug)
            if debug:
                print("Query.optimize: after _RemoveEmpty:----")
                print(optimized.pretty("    "))
            
            #print("Query.optimize(): calling second _SkipLimitApplier: skip=", skip, "   limit=", limit)
            optimized = _SkipLimitApplier().walk(optimized, skip, limit)
            if debug:
                print("Query.optimize: after 2nd _SkipLimitApplier:----")
                print(optimized.pretty("    "))
            
            #optimized = _OrderedApplier()(optimized)
            #if debug:
            #    print("Query.optimize: after applying ordering:----")
            #    print(optimized.pretty("    "))
            
            self.Optimized = optimized
        return self.Optimized

    def compile(self, db=None, skip=0, limit=None, with_meta=False, with_provenance=False, debug=False):
        try:
            optimized = self.optimize(debug=debug, skip=skip, limit=limit)
            optimized = _QueryOptionsApplier().walk(optimized, 
                dict(
                    with_provenance = with_provenance,
                    with_meta = with_meta
                ))
            if debug:
                print("after _QueryOptionsApplier:", optimized.pretty())
            self.Compiled = compiled = SQLConverter(db, debug=debug, include_retired=self.IncludeRetired)(optimized)
        except Exception as e:
            raise MQLCompilationError(traceback.format_exc(limit=-1))

        if debug:
            print("\nCompiled:", compiled.pretty())

        return compiled

    def run(self, db=None, filters={}, skip=0, limit=None, with_meta=True, with_provenance=True, debug=False):

        compiled = self.compile(db=db, 
                    skip=skip, limit=limit, 
                    with_meta=with_meta, with_provenance=with_provenance,
                    debug=debug)
        try:
            result = FileQueryExecutor(db, filters, debug=debug)(compiled)
        except Exception as e:
            raise MQLExecutionError(str(e))
        assert isinstance(result, DBFileSet)
        return result

class _Assembler(Ascender):

    def __init__(self, default_namespace, db=None, loader=None):
        Ascender.__init__(self)
        self.DB = db
        self.Loader = loader
        self.DefaultNamespace = default_namespace
        
    def named_query(self, node, name=None, namespace=None):
        namespace = namespace or self.DefaultNamespace
        if self.DB is not None:
            parsed = MQLQuery.from_db(self.DB, namespace, name)
        elif self.Loader is not None:
            parsed = MQLQuery.from_loader(self.Loader, namespace, name)
        else:
            raise MQLCompilationError(f"Can not load named query {namespace}:{name}")
        assert parsed.Type == "file"
        tree = parsed.Tree
        #tree = _WithParamsApplier().walk(tree, {"namespace":namespace})
        #print("_Assembler.named_query: returning:", tree.pretty())
        return tree
        
class _OrderedApplier(Descender):
    
    def walk(self, tree, ordered=False):
        return Descender.walk(self, tree, ordered)
        
    __call__ = walk

    def ordered(self, node, ordered):
        children = [self.walk(c) for c in node.C]
        return node.clone(children=children)
        
    def basic_file_query(self, node, ordered):
        node["query"].Ordered = node["query"].Ordered or ordered
        return node
        
    def parents_of(self, node, ordered):
        return node.clone(ordered=ordered or node.get("ordered"))

    def children_of(self, node, ordered):
        return node.clone(ordered=ordered or node.get("ordered"))

    basic_dataset_query = basic_file_query

    def skip_limit(self, node, ordered):
        if ordered and node["skip"] == 0:
            child = node.C[0]
            
        child = self.walk(node.C[0], False)
        return node.clone(children=[child])

    def filter(self, node, ordered):
        children = [self.walk(c) for c in node.C]
        node = node.clone(children = children, ordered = node.get("ordered", False) or ordered)
        return node

    def file_list(self, node, ordered):
        return node     # already fixed order

    def _default(self, node, ordered):
        children = [self.walk(c) for c in node.C]
        node = node.clone(children=children)
        if ordered:
            return Node("ordered", [node])
        else:
            return node

class _SkipLimitApplier(Descender):
    #
    # Optimize skip/limit: combine them if in the right order, push them into BFQs when possible
    # skip_limit node is applied by applying "skip" first and then "limit", consistently with Postgres SQL
    #
    
    #def skip(self, node, skip_limit):
    #    skip, limit = skip_limit
    #    node_skip = node["skip"]
    #    return self.walk(node.C[0], (skip+node_skip, limit))
    
    def walk(self, tree, skip=0, limit=None):
        return Descender.walk(self, tree, (skip, limit))
    
    def meta_filter(self, node, skip_limit):
        node = Node("meta_filter", 
            query = self.walk(node["query"]),     # apply skips/limits inside the query
            meta_exp = node["meta_exp"]
        )
        return self._default(node, skip_limit)
    
    def skip_limit(self, node, skip_limit):
        skip, limit = skip_limit
        node_skip = node.get("skip", 0)
        node_limit = node.get("limit")
        #print("_SkipLimitApplier.skip_limit(): node:", node_skip, node_limit, "   context:", skip, limit)
        skip, limit = _merge_skip_limit(node_skip, node_limit, skip, limit)
        #print("           merged:", skip, limit)
        if limit is not None and limit <= 0:
            return Node("empty")
        else:
            applied = self.walk(node.C[0], skip, limit)
            #print("        applied:")
            #print(applied.pretty(indent="        "))
            return applied

    def basic_file_query(self, node, skip_limit):
        query = node["query"]
        #print("_SkipLimitApplier: applying skip_limit", skip_limit, " to BFQ:", query)
        skip, limit = skip_limit
        query.add_skip_limit(skip, limit)
        if query.Limit is not None and query.Limit <= 0:
            return Node("empty")
        else:
            return node

    def union(self, node, skip_limit):
        #print("_SkipLimitApplier: union: skip_limit:", skip_limit, "  children:", node.C)
        skip, limit = skip_limit
        if limit is not None and limit <= 0:
            node = Node("empty")
        else:
            node = Node("union", [self.walk(c) for c in node.C])
            if skip or limit:
                node = Node("skip_limit", [node], skip=skip, limit=limit)
        return node

    def join(self, node, skip_limit):
        #print("_SkipLimitApplier: skip_limit:", skip_limit, "  children:", node.C)
        skip, limit = skip_limit
        if limit is not None and limit <= 0:
            node = Node("empty")
        else:
            node = Node("join", [self.walk(c) for c in node.C])
            if skip or limit:
                node = Node("skip_limit", [node], skip=skip, limit=limit)
        return node

    def filter(self, node, skip_limit):
        skip, limit = skip_limit
        node_skip = node["skip"]
        node_limit = node["limit"]
        skip, limit = _merge_skip_limit(node_skip, node_limit, skip, limit)
        node["limit"] = limit
        node["skip"] = skip
        node.C = [self.walk(c) for c in node.C]
        if limit is not None and limit <= 0:
            return Node("empty")
        else:
            return node

    def file_list(self, node, skip_limit):
        skip, limit = skip_limit
        node_skip = node["skip"]
        node_limit = node["limit"]
        skip, limit = _merge_skip_limit(node_skip, node_limit, skip, limit)
        node["limit"] = limit
        node["skip"] = skip
        if limit is not None and limit <= 0:
            return Node("empty")
        else:
            return node
    
    def empty(self, node, skip_limit):
        return node

    def _default(self, node, skip_limit):
        # print("_LimitApplier._default: node:", node.pretty())
        skip, limit = skip_limit
        node = self.visit_children(node, (0, None))
        if skip or limit:
            node = Node("skip_limit", [node], skip=skip, limit=limit)
        return node

class _RemoveEmpty(Ascender):
    
    def union(self, node, *children):
        children = [c for c in children if c.T != "empty"]
        #print("_RemoveEmpty.union: filtered children:", children)
        if not children:
            return Node("empty")
        elif len(children) == 1:
            #print("         returning:", children[0])
            return children[0]
        else:
            return Node("union", children)

    def join(self, node, *children):
        if any(c.T == "empty" for c in children):
            return Node("empty")
        else:
            return node
            
    def minus(self, node, left, right):
        if right.T == "empty" or left.T == "empty":
            return left
        else:
            return node
            
    def skip_limit(self, node, child, skip=0, limit=None):
        if child.T == "empty":
            return child
        else:
            return node
            
class _QueryOptionsApplier(Descender):
    
    #
    # Applies query params set outside of MQL: with/without metadata, with/without provenance, limit, default namespace
    #
    
    def basic_file_query(self, node, params):
        #print("LimitApplier: applying limit", limit)
        with_meta = params.get("with_meta")
        with_provenance = params.get("with_provenance")
        query = node["query"]
        query.WithMeta = query.WithMeta or with_meta
        query.WithProvenance = query.WithProvenance or with_provenance
        return node
        
    def _default(self, node, params):
        #print("_LimitApplier._default: node:", node.pretty())
        return self.visit_children(node, params)
            
    def meta_filter(self, node, params):
        node["with_provenance"] = params.get("with_provenance", False)
        node["with_meta"] = params.get("with_meta", False)
        new_params = params.copy()
        new_params["with_meta"] = True
        return self.visit_children(node, new_params)

    def parents_of(self, node, params):
        #print("_QueryOptionsApplier.parents_of/children_of: params:", params)
        node["with_provenance"] = params.get("with_provenance", False)
        node["with_meta"] = params.get("with_meta", False)
        new_params = params.copy()
        new_params["with_provenance"] = True
        return self.visit_children(node, new_params)
        
    children_of = parents_of

    def filter(self, node, params):
        # the filter may need metadata but probably not provenance
        new_params = params.copy()
        new_params["with_meta"] = True
        node["with_meta"] = params.get("with_meta", False)
        return self.visit_children(node, new_params)
        
    def file_list(self, node, params):
        node["with_meta"] = node["with_meta"] or params.get("with_meta", False)
        node["with_provenance"] = node["with_provenance"] or params.get("with_provenance", False)
        return node

class _MetaExpPusher(Descender):

    def join(self, node, meta_exp):
        return Node("join", [self.walk(c, meta_exp) for c in node.C])

    def union(self, node, meta_exp):
        return Node("union", [self.walk(c, meta_exp) for c in node.C])
        
    def minus(self, node, meta_exp):
        assert len(node.C) == 2
        left, right = node.C
        return Node("minus", [self.walk(left, meta_exp), self.walk(right, None)])
        
    def basic_file_query(self, node, meta_exp):
        #print("_MetaExpPusher.basic_file_query: meta_exp:", meta_exp)
        if meta_exp is not None:    
            bfq = node["query"]
            assert isinstance(bfq, BasicFileQuery)
            #assert bfq.Relationship is None             # this will be added by ProvenancePusher, as the next step of the optimization
            if bfq.Skip or bfq.Limit:
                node = Node("meta_filter", query=node, meta_exp=meta_exp)
            else:
                bfq.addWhere(meta_exp)
            #bfq.WithMeta = True
        return node
        
    def children_of(self, node, meta_exp):
        if meta_exp is None:
            return self.visit_children(node, None)
        else:
            #
            # meta_filter node is created when we can not push the meta_exp down any further
            #
            return Node("meta_filter", query=self.visit_children(node, None), meta_exp=meta_exp)
        
    parents_of = children_of
    
    def meta_filter(self, node, meta_exp):
        child = node["query"]
        if child.T in ("filter","file_list"):
            return node
        node_exp = node["meta_exp"]
        if node_exp is None:
            new_exp = meta_exp
        elif meta_exp is None:
            new_exp = node_exp
        else:
            new_exp = MetaExpressionDNF.regularize(Node("meta_and", [meta_exp, node_exp]))
        return self.walk(child, new_exp)

class _DatasetEvaluator(Ascender):
    
    def __init__(self, db, with_meta, limit):
        Ascender.__init__(self)
        self.DB = db
        self.WithMeta = with_meta
        self.Limit = limit

    def __call__(self, tree):
        out = self.walk(tree)
        return out

    def dataset_query_list(self, node, *args):
        queries = (a["query"] for a in args)
        return limited(unique(DBDataset.datasets_for_bdqs(self.DB, queries), key=lambda ds: (ds.Namespace, ds.Name)), self.Limit)
        

CMP_OPS = [">" , "<" , ">=" , "<=" , "==" , "=" , "!=", "~~", "~~*", "!~~", "!~~*"]

def _merge_skip_limit(existing_skip, existing_limit, skip=0, limit=None):
    if existing_limit is None:
        return existing_skip+skip, limit
    elif limit is None:
        return existing_skip + skip, max(0, existing_limit - skip)
    else:
        return existing_skip + skip, max(0, min(existing_limit - skip, limit))

class BasicDatasetQuery(object):

    def __init__(self, namespace, name, pattern=False, regexp=False, with_children=False, recursively=False, where=None):
        self.Namespace = namespace
        self.Name = name
        self.Pattern = pattern
        self.RegExp = regexp
        self.WithChildren = with_children
        self.Recursively = recursively
        self.Where = where
        self.Ordered = False
        
    def is_explicit(self):
        return not self.Pattern and not self.WithChildren

    def line(self):
        return "BasicDatasetQuery(%s:%s%s%s%s%s)" % (
                self.Namespace, self.Name, 
                " (pattern) " if self.Pattern and not self.RegExp else (" (pattern re) " if self.Pattern and self.RegExp else ""),
                " with children" if self.WithChildren else "",
                " recursively" if self.Recursively else "",
                (" having:" + self.Where.pretty("    ") if self.Where is not None else ""))

    __str__ = line
    __repr__ = line
                
    def setWhere(self, where):
        self.Where = where
        
    def datasets(self, db, limit=None):
        return DBDataset.datasets_for_bdq(db, self, limit)
        
    def apply_params(self, params):
        # apply params from "with ..."
        default_namespace = params.get("namespace")
        if default_namespace:
            self.Namespace = self.Namespace or default_namespace
                
    def filter_by_where(self, datasets):
        if self.Where is None:
            yield from datasets
        evaluator = MetaEvaluator()
        for ds in datasets:
            if evaluator(ds.Metadata, self.Where):
                yield ds
                
class BasicFileQuery(object):
    
    def __init__(self, dataset_selectors, where=None):
        assert dataset_selectors is None or isinstance(dataset_selectors, list)
        if isinstance(dataset_selectors, list):
            assert all(isinstance(ds, BasicDatasetQuery) for ds in dataset_selectors)
        self.DatasetSelectors = dataset_selectors
        self.Wheres = where
        self.Limit = None
        self.WithMeta = False       
        self.WithProvenance = False
        self.Skip = 0
        self.Ordered = False
        
    def __str__(self):
        out = "BasicFileQuery(datasets:%s, limit:%s, skip:%s, %smeta, %sprovenance%s)" % (
            ",".join(str(s) for s in self.DatasetSelectors or []), 
            self.Limit, self.Skip,
            "with " if self.WithMeta else "no ",
            "with " if self.WithProvenance else "no ",
            "" if not self.Ordered else ", ordered",
            )
        if self.Wheres:
            out += "\n" + self.Wheres.pretty(indent="    ")
        return out

    __repr__ = __str__
        
    def _pretty(self, indent="", headline_indent=""):
        #print(f"BasicFileQuery._pretty(indent='{indent}', headline_indent='{headline_indent}')")
        head = str(self)
        lines = []
        if self.Wheres is not None:
            where_head, where_lines = self.Wheres._pretty(indent + "|   ")
            lines.append(indent + "where = " + where_head)
            lines += where_lines
        return head, lines

    def _pretty(self, indent="", headline_indent=""):
        #print(f"BasicFileQuery._pretty(indent='{indent}', headline_indent='{headline_indent}')")
        head = str(self)
        lines = []
        if self.Wheres is not None:
            where_head, where_lines = self.Wheres._pretty(indent + "|   ")
            lines.append(indent + "where = " + where_head)
            lines += where_lines
        return head, lines

    def pretty(self, indent=""):
        #print("pretty---")
        return "\n".join(self._pretty(indent))

    def addWhere(self, where):
        #print("BasicFileQuery.addWhere():----")
        #print("  self.Wheres:", self.Wheres)
        #print("  where:\n", where.pretty("    "))
        assert isinstance(where, Node) and where.T == "meta_or"
        if self.Wheres is None:
            wheres = where
        else:
            wheres = Node("meta_and", self.Wheres, where)
        self.Wheres = MetaExpressionDNF.regularize(wheres)
        #print("BasicFileQuery.addWhere() result:")
        #print(self.Wheres.pretty("    "))
            
    def addLimit(self, limit):
        self.add_skip_limit(0, limit)

    def addSkip(self, nskip):
        self.add_skip_limit(nskip, None)
            
    def add_skip_limit(self, skip, limit):
        self.Skip, self.Limit = _merge_skip_limit(self.Skip, self.Limit, skip, limit)

    def apply_params(self, params):
        default_namespace = params.get("namespace")
        if self.DatasetSelectors:
            for ds in self.DatasetSelectors:
                ds.apply_params(params)
                
class Orderer(Descender):

    def __call__(self, tree):
        #print("Orderer: input: ", tree.pretty())
        out = self.walk(tree)
        #print("Orderer: output:", out.pretty())
        return out

    def basic_file_query(self, node, _):
        q = node["query"]
        q.Ordered = True
        return node

    basic_dataset_query = basic_file_query
    
    def filter(self, node, _):
        return node.clone(ordered = True)

    parents_of = children_of = filter
    
    def _done(self, node, _):
        return node
        
    ordered = file_list = _done

    def skip_limit(self, node, _):
        if node["skip"]:
            return node     # already ordered
        else:
            node.clone(children=[self(node.C[0])])

    def _default(self, node, _):
        return Node("ordered", [node])

class QueryConverter(Converter):
    
    #
    # converts parsed query (eiher file or dataset) from Lark tree structure to my own Node
    #

    def __init__(self, db=None, loader=None, default_namespace=None):
        self.DB = db
        self.Loader = loader
        self.DefaultNamespace = default_namespace

    def convert(self, tree):
        q = self.transform(tree)
        q.Parsed = tree
        return q
        
    def query(self, args):
        if False and len(args) == 2:
            params, query = args
            # ----  FIXME q = _WithParamsApplier().walk(query, params)
            #print("_Converter.query(): after applying params:", q.pretty())
        else:
            q = args[0]

        if q.T == "top_file_query":         out = FileQuery(q.C[0])
        elif q.T == "top_dataset_query":    out = DatasetQuery(q.C[0])
        else:
            raise ValueError("Unrecognozed top level node type: %s" % (q.T,))
        return out
    
    def __default__(self, typ, children, meta):
        return Node(typ, children, _meta=meta)
        
    def file_query_term(self, args):
        return args[0]

    def merge_meta(self, m1, m2):
        if m1 is None or m2 is None:
            return m1 or m2
        else:
            m = {}
            m.update(m1)
            m.update(m2)
            return m

    def limited_file_query_expression(self, args):
        assert len(args) == 2
        child, limit = args
        limit = int(limit)
        if child.T == "filter":
            child["limit"] = limit
        return Node("limit", [child], limit = limit)
        #return Node("file_query", [args[0]], meta = {"limit":int(args[1].value)})

    def ____make_ordered(self, node):
        if node.T in ("basic_file_query", "basic_dataset_query"):
            q = node["query"]
            q.Ordered = True
        elif node.T in ("filter", "parents_of", "children_of"):
            return node.clone(ordered = True)
        elif node.T in ("file_list", "ordered"):
            pass
        elif node.T == "skip_limit":
            if node["skip"]:
                pass        # aready ordered
            else:
                node = node.clone(children=[self.make_ordered(node.C[0])])
        else:
            node = Node("ordered", [node])
        return node
        
    def make_ordered(self, node):
        return Orderer()(node)

    def ordered(self, args):
        return self.make_ordered(args[0])

    def skip(self, args):
        assert len(args) == 2
        child, skip = args
        skip=int(skip)
        #print("skip: before make_ordered:", child.pretty())
        child = self.make_ordered(child)            # even if skip=0, still make it ordered
        #print("skip: after  make_ordered:", child.pretty())
        if skip == 0:   return child
        if child.T == "basic_file_query":
            q = child["query"]
            skip, limit = _merge_skip_limit(q.Skip, q.Limit, skip=skip)
            q.Skip = skip
            q.Limit = limit
            return child
        elif child.T == "skip_limit":
            skip, limit = _merge_skip_limit(child["skip"], child["limit"], skip=skip)
            return child.clone(skip=skip, limit=limit)
        else:   
            return Node("skip_limit", [child], skip=skip, limit=None)

    def limit(self, args):
        assert len(args) == 2
        child, limit = args
        limit=int(limit)
        if limit == 0:  return Node("empty")
        elif limit is None: return child
        elif child.T == "basic_file_query":
            q = child["query"]
            skip, limit = _merge_skip_limit(q.Skip, q.Limit, limit=limit)
            q.Skip = skip
            q.Limit = limit
            return child
        elif child.T == "skip_limit":
            skip, limit = _merge_skip_limit(child["skip"], child["limit"], limit=limit)
            return child.clone(skip=skip, limit=limit)
        else:
            return Node("skip_limit", [child], limit=limit, skip=0)     # do not make ordered for limit

    def meta_filter(self, args):
        q, meta_exp = args
        meta_exp=MetaExpressionDNF.regularize(meta_exp)
        if q.T == "basic_file_query":
            bfq = q["query"]
            if not (bfq.Skip or bfq.Limit):
                bfq.addWhere(meta_exp)
                return q
        return Node("meta_filter", query=q, meta_exp=meta_exp)

    def basic_file_query(self, args):
        if args:
            #print(args[0].pretty())
            assert len(args) == 1, "Expected 0 or 1 dataset selector list. Got: "+str(args)
            #print("Converter.basic_file_query: args[0].T:", args[0].T)
            assert args[0].T == "dataset_query_list"
            return Node("basic_file_query", query=BasicFileQuery([a["query"] for a in args[0].C]))
        else:
            return Node("basic_file_query", query=BasicFileQuery(None))

    def name_list(self, args):
        return [a.value for a in args]

    def file_list(self, args):
        spec_type = args[0].value
        assert spec_type in ("files", "fids", "file", "fid")
        if spec_type.endswith('s'):
            spec_type = spec_type[:-1]
        if spec_type == "file":
            namespace = self.DefaultNamespace
            specs = []
            for qname in args[-1].C:
                assert qname.T == "qualified_name"
                namespace = qname.get("namespace") or namespace
                specs.append(dict(namespace=namespace, name=qname["name"]))
        else:       # "fids"
            #print(args[-1], args[-1].C)
            specs = [fid.value for fid in args[-1].C]
        return Node("file_list", specs=specs, spec_type=spec_type,
                 with_meta=False, with_provenance=False, limit=None)

    def int_constant(self, args):
        v = args[0]
        return Node("int", value=int(v.value))
        
    def float_constant(self, args):
        v = args[0]
        return Node("float", value=float(v.value))

    def bool_constant(self, args):
        v = args[0]
        #print("bool_constant:", args, args[0].value)
        return Node("bool", value=v.value.lower() == "true")
        
    def unpack_string(self, node):
        assert node.type in ("STRING", "UNQUOTED_STRING")
        s = node.value
        if node.type == "STRING":
            if s[0] in ('"', "'"):
                s = s[1:-1]
        if '"' in s or "'" in s:        # sanitize
            raise ValueError("Unsafe string constant containing double or single quote: %s" % (repr(s),))
        return s

    def datetime_constant(self, args):
        s = self.unpack_string(args[0])
        dt = datetime.fromisoformat(s)
        if dt.tzinfo is None:
            dt = dt.replace(tzinfo=timezone.utc)
        return Node("float", value=float(dt.timestamp()))

    def date_constant(self, args):
        dt = self.unpack_string(args[0])
        date.fromisoformat(dt)         # check if it parses
        if len(args) == 1:
            dt = datetime.fromisoformat(dt)
            dt = dt.replace(tzinfo=timezone.utc)
        else:
            tzshift = self.unpack_string(args[1]) or "+00:00"
            if tzshift[0] not in "-+":
                tzshift = "+" + tzshift
            dt = datetime.fromisoformat(dt + " 00:00:00" + tzshift)
        return Node("date_constant", value=float(dt.timestamp()))

    def string_constant(self, args):
        s = self.unpack_string(args[0])
        return Node("string", value=s)

    #def constant_list(self, args):
    #    return [n["value"] for n in args]

    def qualified_name(self, args):
        assert len(args) in (1,2)
        if len(args) == 1:
            out = Node("qualified_name", namespace=self.DefaultNamespace, name=args[0].value)      # no namespace
        else:
            out = Node("qualified_name", namespace=args[0].value, name=args[1].value)
        #print("Converter.qualified_name: returning: %s" % (out.pretty(),))
        return out
        
    def named_query(self, args):
        if self.DB is None and self.Loader is None:
            raise RuntimeError("Can not load named query without access to the database or query loader")
        (q,) = args
        namespace = q["namespace"] or self.DefaultNamespace
        name = q["name"]
        if self.DB is not None:
            loaded = MQLQuery.from_db(self.DB, namespace, name, convert=False)
        else:
            loaded = MQLQuery.from_loader(self.Loader, namespace, name, convert=False)
        if loaded is None:
            raise ValueError("Named query %s:%s not found" % (namespace, name))
        tree = self.convert(loaded)
        #print("named_query: parsed tree:", tree.pretty())
        if tree.T != "top_file_query":
            raise ValueError(f"Named query {namespace}:{name} must be a file query")
        return tree.C[0]

    def param_def_list(self, args):
        return dict([(a.C[0].value, a.C[1]["value"]) for a in args])
        
    def filter_params(self, args):
        if len(args) == 2:
            assert isinstance(args[0], list)
            assert isinstance(args[1], dict)
            return tuple(args)
        else:
            assert len(args) == 1
            if isinstance(args[0], list):
                return (args[0], {})
            elif isinstance(args[0], dict):
                return ([], args[0])
            else:
                assert False, "Unknown type of filter param:" + str(args[0])

    def union(self, args):
        assert len(args) == 1
        args = args[0].C
        if len(args) == 1:  return args[0]
        unions = []
        others = []
        for a in args:
            if isinstance(a, Node) and a.T == "union":
                unions += a[1:]
            else:
                others.append(a)
        return Node("union", unions + others)
        
    def join(self, args):
        #print("join: args:", args)
        #for a in args:
        #    print("  ", a.pretty())
        assert len(args) == 1
        args = args[0].C
        if len(args) == 1:  return args[0]
        joins = []
        others = []
        for a in args:
            if isinstance(a, Node) and a.T == "join":
                joins += a.C
            else:
                others.append(a)
        return Node("join", joins + others)

    def params_list(self, args):
        # convert date, datetime to floats
        assert len(args) == 1 and args[0].T == "constant_list"
        return [c["value"] for c in args[0].C]

    def filter(self, args):
        if len(args) == 3:
            name, (params, kv), queries = args
        else:
            assert len(args) == 2
            name, queries = args
            params = []
            kv = {}
        queries = queries.C
        for q in queries:
            for bfq in q.find_all("basic_file_query"):
                bfq["query"].WithMeta = True
        node = Node("filter", queries, name = name.value, params=params, kw=kv, skip=0, limit=None, ordered=False, with_meta=False)
        return node

    def meta_attribute(self, args):
        (t,) = args
        return Node("meta_attribute", name=t.value)

    def object_attribute(self, args):
        (t,) = args
        return Node("object_attribute", name=t.value)

    def _convert_array_all(self, node):
        left = node.C[0]
        if left.T == "array_all":
            if node.T == "cmp_op":
                new_op = {
                    "~":   "!~",
                    "!~":  "~",
                    "~*":   "!~*",
                    "!~*":  "~*",
                    ">":    "<=",
                    "<":    ">=",
                    ">=":    "<",
                    "<=":    ">",
                    "=":    "!=",
                    "==":    "!=",
                    "!=":    "=="
                }[node["op"]]
                node["op"] = new_op
            else:
                node.T = {
                    "in_set":"not_in_set",
                    "in_range":"not_in_range",
                    "not_in_set":"in_set",
                    "not_in_range":"in_range",
                }[node.T]
            left.T = "array_any"
            node["neg"] = not node.get("neg")
        #print("_convert_array_all: returning:", node.pretty())
        return node
    
    def array_any(self, args):
        (n,) = args
        return Node("array_any", name=n.value)
        
    def array_all(self, args):
        (n,) = args
        return Node("array_all", name=n.value)
        
    def array_length(self, args):
        (n,) = args
        return Node("array_length", name=n.value)
        
    def array_subscript(self, args):
        name, inx = args
        if inx.type == "STRING":
            inx = inx.value[1:-1]
        else:
            inx = int(inx.value)
        return Node("array_subscript", name=name.value, index=inx)

    def json_path(self, args):
        node = Node("json_path", [args[0]], neg=False)

    def cmp_op(self, args):
        left, op, right = args
        if right.T == "date_constant":
            t = right["value"]
            if op in ("=", "=="):
                node = Node("meta_and",
                    [
                        Node("cmp_op", [left, Node("float", value=t)], op=">=", neg=False),
                        Node("cmp_op", [left, Node("float", value=t + 3600*24)], op="<", neg=False),
                    ]
                )
            elif op == "!=":
                node = Node("meta_or",
                    [
                        Node("cmp_op", [left, Node("float", value=t)], op="<", neg=False),
                        Node("cmp_op", [left, Node("float", value=t + 3600*24)], op=">=", neg=False),
                    ]
                )
            elif op == "<":
                node = Node("cmp_op", [left, Node("float", value=t)], op="<", neg=False)
            elif op == "<=":
                node = Node("cmp_op", [left, Node("float", value=t + 3600*24)], op="<", neg=False)
            elif op == ">":
                node = Node("cmp_op", [left, Node("float", value=t + 3600*24)], op=">=", neg=False)
            elif op == ">=":
                node = Node("cmp_op", [left, Node("float", value=t)], op=">=", neg=False)
            else:
                raise ValueError("Unsopported comparison operation for date constant: %s", op)
        else:
            node = Node("cmp_op", [args[0], args[2]], op=args[1].value, neg=False)
        return self._convert_array_all(node)
        
    def constant_in_array(self, args):
        return Node("cmp_op",
            [Node("array_any", name=args[1].value), args[0]], op="=", neg=False
        )
        
    def constant_not_in_array(self, args):
        return Node("cmp_op",
            [Node("array_any", name=args[1].value), args[0]], op="=", neg=True
        )
        
    def constant_in(self, args):
        const_arg = args[0]
        const_type = const_arg.T
        const_value = const_arg["value"]
        array_in = Node("cmp_op", [Node("array_any", name=args[1].value), const_arg], op="=", neg=False)
        if const_type == "string":
            return Node("meta_or",
                [   array_in,
                    Node("cmp_op", [
                        Node("meta_attribute", name=args[1].value), 
                        Node("string", value=".*%s.*" % (const_value,))
                    ], op="~", neg=False)
                ]
            )
        else:
            return array_in
            
        
    def constant_not_in(self, args):
        const_arg = args[0]
        const_type = const_arg.T
        const_value = const_arg["value"]
        array_not_in = Node("cmp_op", [Node("array_any", name=args[1].value), const_arg], op="=", neg=True)
        if const_type == "string":
            return Node("meta_and",
                [   array_not_in,
                    Node("cmp_op", [
                        Node("meta_attribute", name=args[1].value), 
                        Node("string", value=".*%s.*" % (const_value,))
                    ], op="~", neg=True)
                ]
            )
        else:
            return array_not_in
        
    def in_range(self, args):
        assert len(args) == 3 and args[1].T in ("string", "int", "float", "date_constant") and args[2].T in ("string", "int", "float", "date_constant")
        assert args[1].T == args[2].T, "Range ends must be of the same type"
        typ = args[1].T
        if typ == "date_constant":
            left = args[0]
            low = args[1]["value"]
            high = args[2]["value"] + 24*3600
            node = Node("meta_and",
                [
                    Node("cmp_op", [left, Node("float", value=low)], op=">="),
                    Node("cmp_op", [left, Node("float", value=high)], op="<"),
                ]
            )
        else:
            node = Node("in_range", [args[0]], low=args[1]["value"], high=args[2]["value"], neg=False, type=args[1].T)
        return self._convert_array_all(node)
    
    def not_in_range(self, args):
        assert len(args) == 3 and args[1].T in ("string", "int", "float", "date_constant") and args[2].T in ("string", "int", "float", "date_constant")
        assert args[1].T == args[2].T, "Range ends must be of the same type"
        typ = args[1].T
        if typ == "date_constant":
            left = args[0]
            low = args[1]["value"]
            high = args[2]["value"] + 24*3600
            node = Node("meta_or",
                [
                    Node("cmp_op", [left, Node("float", value=low)], op="<"),
                    Node("cmp_op", [left, Node("float", value=high)], op=">="),
                ]
            )
        else:
            node = Node("in_range", [args[0]], low=args[1]["value"], high=args[2]["value"], neg=True, type=args[1].T)
        return self._convert_array_all(node)

    def in_set(self, args):
        assert len(args) == 2 and args[1].T == "constant_list"
        constant_list = args[1]
        if any(c.T == "date_constant" for c in constant_list.C):
            raise ValueError("in_set operation is not supported for date()")
        values = [c["value"] for c in constant_list.C]
        return self._convert_array_all(Node("in_set", [args[0]], neg=False, set=values))
        
    def not_in_set(self, args):
        assert len(args) == 2 and args[1].T == "constant_list"
        constant_list = args[1]
        if any(c.T == "date_constant" for c in constant_list.C):
            raise ValueError("in_set operation is not supported for date()")
        values = [c["value"] for c in constant_list.C]
        return self._convert_array_all(Node("in_set", [args[0]], neg=True, set=values))
        
    def index(self, args):
        return args[0].value
        
    def meta_and(self, args):
        children = []
        for a in args:
            if a.T == "meta_and":
                children += a.C
            else:
                children.append(a)
        return Node("meta_and", children)
        
    def meta_or(self, args):
        children = []
        for a in args:
            if a.T == "meta_or":
                children += a.C
            else:
                children.append(a)
        return Node("meta_or", children)
        
    def present(self, args):
        assert len(args) == 1
        return Node("present", name = args[0].value)

    def not_present(self, args):
        assert len(args) == 1
        return Node("not_present", name = args[0].value)

    def _apply_not(self, node):
        
        def reverse_array_wildcard(node):
            if node.T == "array_any":
                node = node.clone()
                node.T = "array_all"
            elif node.T == "array_all":
                node = node.clone()
                node.T = "node_any"
            else:
                pass
            return node
        
        if node.T in ("meta_and", "meta_or") and len(node.C) == 1:
            return self._apply_not(node.C[0])
        if node.T == "meta_and":
            return Node("meta_or", [self._apply_not(c) for c in node.C])
        elif node.T == "meta_or":
            return Node("meta_and", [self._apply_not(c) for c in node.C])
        elif node.T == "meta_not":
            return node.C[0]
        elif node.T in ("cmp_op", "in_set", "in_range", "json_path"):   # why cmp_op is here ??? 
            node["neg"] = not node.get("neg")
            return node
        elif node.T == "cmp_op":                                        # why cmp_op is not here ???
            new_op = {
                "~":   "!~",
                "!~":  "~",
                "~*":   "!~*",
                "!~*":  "~*",
                ">":    "<=",
                "<":    ">=",
                ">=":    "<",
                "<=":    ">",
                "=":    "!=",
                "==":    "!=",
                "!=":    "=="
            }[node["op"]]
            return node.clone(op=new_op)
        elif node.T == "present":
            return Node("not_present", name=node["name"])
        elif node.T == "not_present":
            return Node("present", name=node["name"])
        else:
            raise ValueError("Unknown node type %s while trying to apply NOT operation" % (node.T,))
            
    def meta_not(self, children):
        assert len(children) == 1
        return self._apply_not(children[0])
        
    #
    # Datasets
    #
    
    def _______did_pattern(self, args):
        assert len(args) in (1,2)
        namespace = self.DefaultNamespace
        if len(args) == 1:
            name = args[0].value
        else:
            namespace, name = args[0].value, args[1].value
        # unquote the string
        if name.startswith("'") or name.startswith('"'):
            name = name[1:-1]
        name = name.replace('*', '%')
        name = name.replace('?', '_')
        out = Node("did_pattern", namespace=namespace, name=name)
        return out

    def regexp_pattern(self, args):
        assert len(args) in (1,2)
        namespace = self.DefaultNamespace
        if len(args) == 1:
            name = args[0].value
        else:
            namespace, name = args[0].value, args[1].value
        name = name[1:-1]           # remove enclosing quotes
        out = Node("did_pattern", namespace=namespace, name=name)
        return out

    def sql_pattern(self, args):
        assert len(args) in (1,2)
        namespace = self.DefaultNamespace
        if len(args) == 1:
            name = args[0].value
        else:
            namespace, name = args[0].value, args[1].value
        name = name.replace("*", "%").replace("?", "_")
        out = Node("did_pattern", namespace=namespace, name=name)
        return out

    def dataset_spec(self, args):
        name_or_pattern = args[-1]
        pattern = "matching" in args
        regexp = "regexp" in args
        assert name_or_pattern.T in ("did_pattern", "qualified_name")
        name = name_or_pattern["name"]
        namespace = name_or_pattern.get("namespace") or self.DefaultNamespace
        query = BasicDatasetQuery(namespace, name, pattern=pattern, regexp=regexp)
        return Node("basic_dataset_query", query=query)
        
    def dataset_add_where(self, children):
        assert len(children) == 2
        bdq, where = children
        assert bdq.T == "basic_dataset_query", "Unknown node type: "+bdq.pretty()
        q = bdq["query"]
        q.setWhere(children[1])
        return Node("basic_dataset_query", query=q)

    def dataset_add_subsets(self, children):
        assert len(children) == 2
        bdq, subsets = children
        q = bdq["query"]
        q.WithChildren = True
        q.Recursively = subsets["recursive"]
        return Node("basic_dataset_query", query=q)

    def dataset_provenance_op(self, children):
        return Node("subsets", recursive=any(c.value == "recursively" for c in children))

class MQLQuery(object):
    
    @staticmethod
    def parse(text, db=None, loader=None, debug=False, convert=True, default_namespace=None, include_retired_files=None):
        out = []
        for l in text.split("\n"):
            l = l.split('#', 1)[0]
            out.append(l)
        text = '\n'.join(out)
        try:
            parsed = _Parser.parse(text)
            if debug:
                print("parsed:\n", parsed.pretty())
            if convert:
                converted = QueryConverter(db=db, loader=loader, default_namespace=default_namespace).convert(parsed)
                if debug:
                    print("converted:\n", converted.pretty())
                if converted.T == "top_file_query":
                    q = FileQuery(converted.C[0], include_retired_files)
                else:
                    q = DatasetQuery(converted.C[0])
                q.Parsed = parsed
                return q
            else:
                return parsed
        except LarkError as e:
            raise MQLSyntaxError(str(e))
        
    @staticmethod
    def from_db(db, namespace, name, convert=True):
        q = DBNamedQuery.get(db, namespace, name)
        if q is None:
            raise ValueError("Named query %s:%s not found" % (namespace, name))
        text = q.Source
        return MQLQuery.parse(text, convert=convert)
        
    @staticmethod
    def from_loader(loader, namespace, name, convert=True):
        text = loader.get_named_query(namespace, name)
        if text is None:
            raise ValueError("Named query %s:%s not found" % (namespace, name))
        return MQLQuery.parse(text, convert=convert)
        
    @staticmethod
    def from_db_raw(db, namespace, name):
        q = DBNamedQuery.get(db, namespace, name)
        if q is None:
            raise ValueError("Named query %s:%s not found" % (namespace, name))
        return q.Source



FileQuery = """

top_file_query          :    file_query

?file_query: meta_filter                                  
    | file_query "-" meta_filter                          -> minus

?meta_filter: file_query_exression "where" meta_exp     
    |   file_query_exression                             

?file_query_exression:  file_query_term                   
    |   "union" "(" file_query_list ")"                  -> union
    |   "[" file_query_list "]"                          -> union
    |   "join"  "(" file_query_list ")"                  -> join
    |   "{" file_query_list "}"                          -> join
    |   "parents" "(" file_query ")"                     -> parents_of
    |   "children" "(" file_query ")"                    -> children_of
    |   file_query "limit" SIGNED_INT                    -> limit              
    |   file_query "skip" SIGNED_INT                     -> skip
    |   file_query "ordered"                             -> ordered
    |   "(" file_query ")"           

file_query_term: "files" ("from" "datasets"? dataset_query_list)?                               -> basic_file_query
    |   "filter" FNAME "(" filter_params ? ")" "(" file_query_list ")"              -> filter
    |   "query" qualified_name                                                      -> named_query
    |   file_list

!file_list: ("fids"|"fid") fid_list
    |   ("files"|"file") qualified_name_list

filter_params : params_list
    |   (params_list ",")? param_def_list

params_list : constant_list         // convert date, datetime to floats

file_query_list: file_query ("," file_query)*     

// file attributes
FILE_ATTR_NAME: ("id" | "namespace" | "name" | "creator" | "updated_by" | "created_timestamp" | "updated_timestamp" | "retired" | "retired_by" | "retired_timestamp" )

"""


import sys, getopt, os, json, fnmatch, pprint
from urllib.parse import quote_plus, unquote_plus
from metacat.util import to_bytes, to_str
from metacat.webapi import MetaCatClient, MCError
from metacat.ui.cli import CLI, CLICommand, InvalidOptions, InvalidArguments

Usage = """
Usage: 
    metacat namespce <command> [<options>] ...
    
    Commands and options:
    
        list [<options>] [<namespace pattern>]
            -v -- verbose
            
        create [<options>] <name>
            -o <role> -- owner role
        
        show <name>
"""

class ListCommand(CLICommand):

    GNUStyle = True
    Opts = "dvu:r: verbose user= role= directly"
    Usage = """[options] [<pattern>]
        <pattern> is a UNIX shell style pattern (*?[]), optional
        -u|--user <username>        - list namespaces owned by the user
        -d                          - exclude namespaces owned by the user via a role
        -r|--role <role>            - list namespaces owned by the role
    """

    
    def __call__(self, command, client, opts, args):
        pattern = None if not args else args[0]
    
        opts = dict(opts)
        verbose = "-v" in opts or "--verbose" in opts
        match_owner_user = opts.get("-u", opts.get("--user"))
        match_owner_role = opts.get("-r", opts.get("--role"))
        if match_owner_user and match_owner_role:
            raise InvalidOptions("Owner user and owner role can not be used together")
        output = client.list_namespaces(pattern=pattern, owner_user=match_owner_user, owner_role=match_owner_role, directly="-d" in opts)
        for item in output:
            name = item["name"]
            owner_user = item.get("owner_user")
            owner_role = item.get("owner_role")
            if owner_user:
                owner = "u:"+owner_user
            else:
                owner = "r:"+owner_role
            print("%-30s\t%-20s\t%s" % (name, owner, item.get("descrition") or ""))
                
class ShowCommand(CLICommand):
    
    GNUStyle = True
    Opts = "j json"
    MinArgs = 1
    Usage = """[-j|--json] <namespace>
        -j|--json           - print as JSON
    """
    
    
    def __call__(self, command, client, opts, args):
        data = client.get_namespace(args[0])
        if "-j" in opts or "--json" in opts:
            print(json.dumps(data, indent=4, sort_keys=True))
        else:
            pprint.pprint(data)

class CreateCommand(CLICommand):

    GNUStyle = True
    Opts = "o:j json owner="
    MinArgs = 1
    Usage = """[options] <namespace> [<description>]
        -o <owner>|--owner <owner>              - namespace owner
        -j|--json                               - print as JSON 
    """
    
    
    def __call__(self, command, client, opts, args):
        name = args[0]
        description = (" ".join(args[1:])).strip() or None
        data = client.create_namespace(name, owner_role=opts.get("-o", opts.get("--owner")), description=description)

        if "-j" in opts or "--json" in opts:
            print(json.dumps(data, indent=4, sort_keys=True))
    
NamespaceCLI = CLI(
    "create",   CreateCommand(),
    "list",     ListCommand(),
    "show",     ShowCommand()
)
 



from .py3 import PY2, PY3, to_str, to_bytes
from .timelib import epoch
from .trace import Tracer
from .object_spec import ObjectSpec, undid
from .utils import first_not_empty, insert_sql
from .validation import validate_metadata
from .generators import fetch_generator, chunked, limited, unique, strided, skipped


import re, traceback
from metacat.common import FileAttributes

class MetaEvaluator(object):

    BOOL_OPS = ("and", "or", "not")
    
    def file_attribute(self, f, attrname):
        x = None
        if attrname == "creator":               x = f.Creator
        elif attrname == "created_timestamp":   x = f.CreatedTimestamp.epoch()
        elif attrname == "name":                x = f.Name
        elif attrname == "namespace":           x = f.Namespace
        elif attrname == "size":                x = f.Size
        return x

    def evaluate_meta_expression(self, f, meta_expression):
        metadata = f.metadata()
        #print("    meta:", metadata)
        op, args = meta_expression.T, meta_expression.C
        if op in ("meta_and", "meta_or") and len(args) == 1:
            return self.evaluate_meta_expression(f, args[0])
        #if meta_expression["neg"]:
        #    return not self.evaluate_meta_expression(metadata, meta_expression.clone(neg=False))
        if op == "meta_and":    op = "and"
        if op == "meta_or":     op = "or"
        if op in self.BOOL_OPS:
            return self.eval_meta_bool(metadata, op, args)
        elif op == "present":
            return meta_expression["name"] in metadata
        elif op == "not_present":
            return not meta_expression["name"] in metadata
        elif op in ("in_set", "not_in_set"):
            neg = meta_expression.get("neg", False) != (op == "not_in_set")
            vset = set(meta_expression.get("set", []))
            left = args[0]
            aname = left["name"]
            if left.T == "scalar":
                v = metadata.get(aname)
                result = v in vset
            elif left.T == "object_attribute":
                v = self.file_attribute(f, aname)
                result = v in vset
            elif left.T == "array_any":
                if not aname in metadata:  return neg
                lst = metadata[aname]
                if not isinstance(lst, list):   return neg
                for x in lst:
                    if x in vset:  
                        result = True
                        break
                else:
                    result = False
            elif left.T == "array_subscript":
                inx = left["index"]
                if not aname in metadata:  return neg
                lst = metadata[aname]
                try:    v = lst[inx]
                except: result = False
                else: result = v in vset
            elif left.T == "array_length":
                if not aname in metadata:  return neg
                lst = metadata[aname]
                if not isinstance(lst, list):
                    result = False
                else:
                    result = len(lst) in vset
            return result != neg        # negate if neg
        elif op in ("in_range", "not_in_range"):
            low, high = meta_expression["low"], meta_expression["high"]
            left = args[0]
            neg = meta_expression.get("neg", False) != (op == "not_in_range")
            if left.T == "scalar":
                aname = left["name"]
                try:    return (aname in metadata and metadata[aname] >= low and metadata[aname] <= high) != neg
                except: return neg
            elif left.T == "array_any":
                aname = left["name"]
                if not aname in metadata:  return neg
                lst = metadata[aname]
                if isinstance(lst, dict):
                    attr_values = lst.values()
                elif isinstance(lst, list):
                    attr_values = lst
                else:
                    return neg
                for x in attr_values:
                    if x >= low and x <= high:  return not neg
                else:
                    return neg
            elif left.T == "array_subscript":
                aname = left["name"]
                inx = left["index"]
                if not aname in metadata:  return neg
                lst = metadata[aname]
                try:    v = lst[inx]
                except: return neg
                return (v >= low and v <= high) != neg                    
            elif left.T == "array_length":
                aname = left["name"]
                if not aname in metadata:  return neg
                lst = metadata[aname]
                if not isinstance(lst, list):
                    return neg
                l = len(lst)
                return (l >= low and l <= high) != neg
        elif op == "cmp_op":
            cmp_op = meta_expression["op"]
            left, right = args
            aname = left["name"]
            value = right["value"]
            if left.T == "meta_attribute":
                try:    
                    result = aname in metadata and self.do_cmp_op(metadata[aname], cmp_op, value)
                    #print("result:", result)
                    return result
                except: return False
            elif left.T == "object_attribute" and aname in FileAttributes:
                try:    
                    v = self.file_attribute(f, aname)
                    result = self.do_cmp_op(self.file_attribute(f, aname), cmp_op, value)
                    return result
                except: 
                    traceback.print_exc()
                    return False
            elif left.T == "array_any":
                aname = left["name"]
                lst = metadata.get(aname)
                #print("lst:", lst)
                if lst is None:  return False
                if isinstance(lst, dict):
                    attr_values = lst.values()
                elif isinstance(lst, list):
                    attr_values = lst
                else:
                    return False
                for av in attr_values:
                    #print("comparing", av, cmp_op, value)
                    if self.do_cmp_op(av, cmp_op, value):
                        return True
                else:
                    return False
            elif left.T == "array_subscript":
                aname = left["name"]
                inx = left["index"]
                lst = metadata.get(aname)
                if lst is None:  return False
                try:    av = lst[inx]
                except: return False
                return  self.do_cmp_op(av, cmp_op, value)                
            elif left.T == "array_length":
                aname = left["name"]
                lst = metadata.get(aname)
                if lst is None:  return False
                if not isinstance(lst, list):
                    return False
                l = len(lst)
                result = self.do_cmp_op(l, cmp_op, value)  
                return result
        raise ValueError("Invalid expression:\n"+meta_expression.pretty())
        
    __call__ = evaluate_meta_expression

    def eval_meta_bool(self, f, bool_op, parts):
        assert len(parts) > 0
        p0 = parts[0]
        rest = parts[1:]
        ok = self.evaluate_meta_expression(f, p0)
        if bool_op in ("and", "meta_and"):
            if len(rest) and ok:
                ok = self.eval_meta_bool(f, bool_op, rest)
            return ok
        elif bool_op in ("or", "meta_or"):
            if len(rest) and not ok:
                ok = self.eval_meta_bool(f, bool_op, rest)
            return ok
        elif bool_op == "not":
            assert len(rest) == 0
            return not ok
        else:
            raise ValueError("Unrecognized boolean operation '%s'" % (bool_op,))
    
    def do_cmp_op(self, x, op, y):
        if op == "<":          return x < y
        elif op == ">":    
            #print("evaluate_meta_expression: > :", attr_value, value)    
            return x > y
        elif op == "<=":       return x <= y
        elif op == ">=":       return x >= y
        elif op in ("==",'='): 
            #print("evaluate_meta_expression:", repr(attr_value), repr(value))
            return x == y
        elif op == "!=":       return x != y
        # - fix elif op == "in":       return value in attr_value       # exception, e.g.   123 in event_list
        elif op in ("~", "!~", "~*", "!~*"):
            negated = op[0] == '!'
            flags = re.IGNORECASE if op[-1] == '*' else 0
            r = re.compile(y, flags)
            match = r.search(x) is not None
            return negated != match
        else:
            raise ValueError("Invalid comparison operator '%s'" % (op,))
            
    def do_cmp_attribute(self, f, attrname, op, y):

        x = None
        if attrname == "creator":               x = f.Creator
        elif attrname == "created_timestamp":   x = f.CreatedTimestamp.epoch()
        elif attrname == "name":                x = f.Name
        elif attrname == "namespace":           x = f.Namespace
        elif attrname == "size":                x = f.Size
        else:
            raise ValueError("Unknown file attribute:", attrname)

        if op == "<":          return x < y
        elif op == ">":    
            #print("evaluate_meta_expression: > :", attr_value, value)    
            return x > y
        elif op == "<=":       return x <= y
        elif op == ">=":       return x >= y
        elif op in ("==",'='): 
            #print("evaluate_meta_expression:", repr(attr_value), repr(value))
            return x == y
        elif op == "!=":       return x != y
        # - fix elif op == "in":       return value in attr_value       # exception, e.g.   123 in event_list
        elif op in ("~", "!~", "~*", "!~*"):
            negated = op[0] == '!'
            flags = re.IGNORECASE if op[-1] == '*' else 0
            r = re.compile(y, flags)
            match = r.search(x) is not None
            return negated != match
        else:
            raise ValueError("Invalid comparison operator '%s'" % (op,))
        
        
    @staticmethod
    def evaluate(meta, exp):
        return MetaEvaluator().evaluate_meta_expression(meta, exp)
    



import traceback, sys, time
from datetime import datetime
from .log_file import LogFile, LogStream, log_writer

DefaultLogger = None

class LogChannel(object):
    
    def __init__(self, output, label=None, enabled=True, timestamps=True):
        assert output is not None
        self.Timestamps = timestamps
        self.Writer = log_writer(output)
        self.Label = label
        self.Enabled = enabled
        
    def enable(self, enabled=True):
        self.Enabled = enabled

    def log(self, who, *message, sep=" ", t=None, label=None):
        #print("LogChannel.log(): who:", who)
        if self.Enabled:
            message = sep.join([str(p) for p in message])
            label = label or self.Label
            if label is not None:
                message = f"[{label}] {message}"
            if who:
                message = f"{who}: {message}"
            if not self.Timestamps: t = False
            self.Writer.log(message, t=t)

class AbstractLogger(object):

    def log(self, *message, sep=" ", who=None, t=None, channel="log"):
        raise NotImplementedError()


class Logger(AbstractLogger):

    def __init__(self, log_path, error_path=None, debug_path=None, debug=True, append=True):
        self.Debug = debug
        writer = log_writer(log_path, append=append)
        
        # default channels
        self.Channels = {       
            "log":      LogChannel(writer),
            "error":    LogChannel(writer if error_path is None else log_writer(error_path, append=append), label="ERROR")
        }
        if debug:
            self.Channels["debug"] = LogChannel(writer if debug_path is None else log_writer(debug_path, append=append), label="DEBUG")

    def add_channel(self, name, path=None, print_label=False, timestamps=True, **params):
        if path:    
            channel = LogChannel(log_writer(path, **params), 
                label = name if print_label else None,
                timestamps = timestamps
                )
        else:
            channel = self.Channels["log"]
        self.Channels[name] = channel

    def log(self, *message, sep=" ", who=None, t=None, channel="log"):
        #print("Logger.log(", message, sep, who, t, channel, ")")
        assert who is not None, "Message originator (who) must be specified"
        channel = self.Channels.get(channel)
        if channel is not None:
            channel.log(who, *message, sep=sep, t=t)

    def error(self, *message, sep=" ", who=None, t=None):
        assert who is not None, "Message originator (who) must be specified"
        self.log(*message, channel="error", sep=sep, who=who, t=t)

    def debug(self, *message, sep=" ", who=None, t=None):
        assert who is not None, "Message originator (who) must be specified"
        if self.Debug:
            self.log(*message, channel="debug", sep=sep, who=who, t=t)

class Logged(AbstractLogger):

    def __init__(self, name=None, debug=True, logger=None,
            log_channel="log", error_channel="error", debug_channel="debug"):
        assert logger is None or isinstance(logger, AbstractLogger), "logger must be either None or a Logger or a Logged"
        self.Logger = logger
        self.LogName = name or self.__class__.__name__
        self.Debug = debug
        self.LogChannel = log_channel
        self.ErrorChannel = error_channel
        self.DebugChannel = debug_channel

    def log(self, *message, sep=" ", who=None, t=None, channel=None):
        #print("Logged.log(", message, sep, who, t, channel, ")")
        channel = channel or self.LogChannel
        who = who or self.LogName
        logger = self.Logger or DefaultLogger
        if logger is not None and (channel != self.DebugChannel or self.Debug):
           logger.log(*message, sep=sep, who=who, t=t, channel=channel)
    
    def error(self, *message, sep=" ", who=None, t=None):
        self.log(*message, sep=sep, who=who or self.LogName, t=t, channel=self.ErrorChannel)

    def debug(self, *message, sep=" ", who=None, t=None):
        self.log(*message, sep=sep, who=who or self.LogName, t=t, channel=self.DebugChannel)


def init(log_output, error_out=None, debug_out=None, debug_enabled=False):
    global DefaultLogger
    DefaultLogger = Logger(log_output, error_out, debug_out, debug_enabled)
    return DefaultLogger
    



from .webapi import MetaCatClient, MCError
from .webapi import ServerReportedError as MCServerError
from .webapi import AuthenticationError
from .webapi import WebAPIError as MCWebAPIError
from .webapi import InvalidMetadataError as MCInvalidMetadataError


import sys, getopt, os, json, pprint
from urllib.request import urlopen, Request
from urllib.parse import quote_plus, unquote_plus
from metacat.util import to_bytes, to_str
from metacat.webapi import MetaCatClient, MCServerError, MCWebAPIError, MCError
from metacat.ui.cli import CLICommand, InvalidArguments, InvalidOptions

class QueryCommand(CLICommand):

    GNUStyle = False    
    Opts = (
        "jim:N:pq:S:A:lPxrL:U:S:R:Q:2t:s", 
        ["line", "json", "ids", "summary=", "metadata=", "namespace=", "pretty",
            "with-provenance", "save-as=", "add-to=", "explain", "include-retired-files",
            "list=", "source=", "create=", "update=", "run=", "1024", "timeout="
        ]
    )
    Usage = """[<options>] (-q <MQL query file>|"<MQL query>")

        Options:
            -t|--timeout <timeout in seconds>   - request timeout (default 600)
            -j|--json                           - print raw JSON output
            -p|--pretty                         - pretty-print metadata
            -l|--line                           - print all metadata on single line (good for grepping, ignored with -j and -p)
            -i|--ids                            - print file ids instead of names
               --summary (count|keys)           - print only summary information
                                                      count: file count and total size
                                                      keys: list of all top level metadata keys for selected files
            -s                                  - shortcut for --summary count
                 -2|--1024                      - for count, print sizes in KiB, GiB (1024, ...), instead of powers of 1000 (KB, GB, ...)
            -m|--metadata [<field>,...]         - print metadata fields
                                                  overrides --summary
            -m|--metadata all                   - print all metadata fields
                                                  overrides --summary
            -P|--with-provenance                - include provenance information
            -N|--namespace=<default namespace>  - default namespace for the query
            -S|--save-as=<namespace>:<name>     - save files as a new datset
            -A|--add-to=<namespace>:<name>      - add files to an existing dataset
            -r|--include-retired-files          - include retired files into the query results
            
            -x|--explain                        - dp not run the query, show resulting SQL only
    """
    
    def __call__(self, command, client, opts, args):
        namespace = opts.get("-N") or opts.get("--namespace")
        #with_meta = not "--summary" in opts and not "-s" in opts
        with_meta = "-m" in opts or "--metadata" in opts
        with_provenance = "-P" in opts or "--with-provenance" in opts
        keys = opts.get("-m") or opts.get("--metadata") or []
        if keys and keys != "all":    keys = keys.split(",")
        save_as = opts.get("-S") or opts.get("--saves-as")
        add_to = opts.get("-A") or opts.get("--add-to")
        include_retired = "-r" in opts or "--include-retired-files" in opts
        summary = opts.get("--summary", "count" if "-s" in opts else None)
        timeout = int(opts.get("-t", opts.get("--timeout", 600)))
        if args:
            query_text = " ".join(args)
        else:
            query_file = opts.get("-q")
            if not query_file:
                raise InvalidArguments("Query must be specified")
            query_text = to_str(open(query_file, "r").read())
            
        if "-x" in opts or "--explain" in opts:
            from metacat.mql import MQLQuery
            print("---- Query text ----\n%s\n" % (query_text,))
            q = MQLQuery.parse(query_text, loader=client, include_retired_files=include_retired)

            print("---- Parsed ----")
            print(q.Parsed.pretty())
            print("")

            print("---- Converted ----")
            print(q.Tree.pretty("    "))
            print("")
        
            #q.optimize(False)
            #print("---- Optimized ----")
            #print(q.Optimized.pretty("    "))
            #print("")
    
            compiled = q.compile(with_meta=with_meta, with_provenance=with_provenance)
            print("---- Compiled ----")
            print(compiled.pretty("    "))
        else:
            client.Timeout = timeout
            results = client.query(query_text, 
                        namespace=namespace, with_metadata = with_meta, 
                        save_as=save_as, add_to=add_to,
                        with_provenance=with_provenance,
                        include_retired_files=include_retired, summary=summary
            )

            in_line = "-l" in opts or "--line" in opts
            print_format = "json" if ("--json" in opts or "-j" in opts) \
                else ("pprint" if "--pretty" in opts or "-p" in opts else "text")

            #print("response results:", results)
    
            if summary == "count":
                if print_format == "text":
                    nfiles = results["count"]
                    total_size = results["total_size"]
                    print("Files:       ", nfiles)
                    K = 1024 if "-2" in opts else 1000
                    U = "iB" if "-2" in opts else "B"
                    if total_size >= K*K*K*K:
                        unit = "T" + U
                        n = total_size / (K*K*K*K)
                    elif total_size >= K*K*K:
                        unit = "G" + U
                        n = total_size / (K*K*K)
                    elif total_size >= K*K:
                        unit = "M" + U
                        n = total_size / (K*K)
                    elif total_size >= K:
                        unit = "K" + U
                        n = total_size / K
                    else:
                        unit = "B"
                        n = total_size
                    print("Total size:  ", "%d (%.3f %s)" % (total_size, n, unit))
                elif print_format == "json":
                    print(json.dumps(results, indent=4))
                else:
                    pprint.pprint(results)
            elif summary == "keys":
                results = list(results)
                #print(results)
                results = sorted(results)
                if print_format == "text":
                    for k in results:
                        print(k)
                elif print_format == "json":
                    print(json.dumps(results, indent=4))
                else:
                    pprint.pprint(results)
            else:
                if print_format == "json":
                    results = list(results)
                    print(json.dumps(results, sort_keys=True, indent=4, separators=(',', ': ')))
                elif print_format == "pretty":
                    results = list(results)
                    meta = sorted(results, key=lambda x: x["name"])
                    pprint.pprint(meta)
                else:
                    for f in results:
                        meta_lst = []
                        meta_out = ""
                        if with_meta:
                            meta = f["metadata"]
                            klist = sorted(list(meta.keys())) if keys == "all" else keys
                            if in_line:
                                for k in klist:
                                    if k in meta:
                                        meta_lst.append("%s=%s" % (k, repr(meta[k])))
                            else:
                                for k in klist:
                                    if k in meta:
                                        meta_lst.append("%s\t=\t%s" % (k, repr(meta[k])))
                            if meta_lst:
                                if in_line:
                                    meta_out = "\t"+"\t".join(meta_lst)
                                else:
                                    meta_out += "\n    "+"\n    ".join(meta_lst)
                        if "--ids" in opts or "-i" in opts:
                            print("%s%s" % (f["fid"],meta_out))
                        else:
                            print("%s:%s%s" % (f["namespace"], f["name"], meta_out))

QueryInterpreter = QueryCommand()


# common functionality for Auth, GUI and Data servers

from wsdbtools import ConnectionPool
from metacat.util import to_str, to_bytes
from metacat.common import (digest_server, 
    SignedToken, SignedTokenExpiredError, SignedTokenImmatureError, SignedTokenUnacceptedAlgorithmError, 
    SignedTokenSignatureVerificationError
)

from metacat.auth import BaseDBUser as DBUser
import psycopg2, json, time, secrets, traceback, hashlib, pprint, os, yaml

class AuthenticationCore(object):

    """Class encapsulating the client authentication funtionality for a group
    """

    DefaultTokenExpiration = 24*3600*7


    def __init__(self, cfg, group=None):
        self.Cfg = cfg
        db_config = cfg.get("user_database") or cfg["database"]
        connstr = self.connstr(db_config)
        self.UserDB = ConnectionPool(postgres=connstr, max_idle_connections=1, idle_timeout=20)
        self.UserDBSchema = db_config.get("schema")

        self.AuthConfig = cfg.get("authentication")
        self.Group = group or self.AuthConfig.get("group", "metacat")          
        self.Realm = self.AuthConfig.get("realm", self.Group) # will be used by the rfc2617 authentication
        self.TokenIssuer = self.AuthConfig.get("issuer")
        self.SciTokenIssuers = self.AuthConfig.get("sci_token_issuers", [])
        secret = self.AuthConfig.get("secret")
        if secret is None:    
            raise ValueError("Authentication secret not found in the configuration")
            self.TokenSecret = secrets.token_bytes(128)     # used to sign tokens
        else:         
            h = hashlib.sha256()
            h.update(to_bytes(secret))      
            self.TokenSecret = h.digest()
        self.TokenExpiration = self.DefaultTokenExpiration

    def connstr(self, cfg):
        cs = "host=%(host)s port=%(port)s dbname=%(dbname)s user=%(user)s" % cfg
        if cfg.get("password"):
            cs += " password=%(password)s" % cfg
        return cs

    def auth_config(self, method):
        # configuration for particular authentication mechanism
        return self.AuthConfig.get(method)

    def user_db(self):
        conn = self.UserDB.connect()
        if self.UserDBSchema:
            conn.cursor().execute(f"set search_path to {self.UserDBSchema}")
        return conn
        
    def get_user(self, username):
        db = self.user_db()
        return DBUser.get(db, username)

    def verify_token(self, encoded_token):
        try:    
            token = SignedToken.from_bytes(encoded_token)
            #print("verify_token: token:", token.Payload)
            t = time.time()
            #print("   time:", t, "  t-iat:", t-token["iat"], "  t-nbf:", t-token["nbf"])
            token.verify(self.TokenSecret)
            #print("verify_token: verified. subject:", token.subject)
        except SignedTokenExpiredError:
            return None, "Token expired"           
        except SignedTokenImmatureError:
            return None, "Token immature"           
        except SignedTokenUnacceptedAlgorithmError:
            return None, "Invalid token algorithm"           
        except SignedTokenSignatureVerificationError:
            return None, "Token verification failed"           
        except Exception as e:
            return None, str(e)
        else:
            #print("verify_token: token:", token, "  subject:", token.subject)
            return token, None

    def generate_token(self, user, payload={}, expiration=None):
        if expiration is None:
            expiration = self.TokenExpiration
        token = SignedToken(payload, subject=user, expiration=expiration, issuer=self.TokenIssuer)
        #print("generate_token: payload:", token.Payload)
        return token, token.encode(self.TokenSecret)

    def get_digest_password(self, realm, username):
        u = self.get_user(username)
        if u is None:
            return None
        return u.get_password(self.Realm)

    def _auth_digest(self, request_env, redirect):
        # give them cookie with the signed token
        
        ok, data = digest_server(self.Realm, request_env, self.get_digest_password)   # use the Group as realm, if present
        if ok:
            username = data
            return "ok", dict(username=username)
        elif data:
            return "continue", (401, "Authorization required", {
                'WWW-Authenticate': data
            })
        else:
            return "reject", "Authentication failed"

    def _auth_ldap(self, request, redirect, username):
        ldap_config = self.AuthConfig.get("ldap")
        if not ldap_config:
            return "reject", "LDAP is not configured"
        if username:
            password = to_str(request.body.strip())
        else:
            username, password = request.body.split(b":",1)
            username = to_str(username)
            password = to_str(password)
        u = self.get_user(username)
        config = self.AuthConfig["ldap"]
        result, reason, expiration = u.authenticate("ldap", config, password)
        #print("AuthCore._auth_ldap:", result, reason, expiration)
        if result:
            return "ok", dict(username=username)
        else:
            return "reject", "Authentication failed"

    def accepted_sci_token_issuer(self, issuer):
        return issuer in self.SciTokenIssuers

    def _auth_token(self, request, redirect, username):
        u = self.get_user(username)
        if u is None:
            return "reject", "Authentication failed"

        encoded = None
        headers = request.headers
        authorization = headers.get("Authorization")
        if authorization:
            try:
                encoded = authorization.split(None, 1)[-1]      # ignore "type", e.g. bearer
            except:
                pass

        if not encoded:
            encoded = request.cookies.get("auth_token") or request.headers.get("X-Authentication-Token")
        if not encoded:
            return "reject", "Authentication failed. Token not found"

        result, reason, expiration = u.authenticate("scitoken", self.SciTokenIssuers, encoded)
        if not result:
            result, reason, expiration = u.authenticate("jwttoken", 
                    {
                        "issuer": self.TokenIssuer,
                        "secret": self.TokenSecret
                    }, encoded)
        if result:
            return "ok", dict(username=username, expiration=expiration)
        else:
            return "reject", "Authentication failed"

    def _auth_x509(self, request, redirect, username):
        #log = open("/tmp/_auth_x509.log", "w")
        #print("_auth_x509: request.environ:")
        #for k, v in sorted(request.environ.items()):
        #    print(f"   {k}={v}")
        #print("_auth_x509: scheme:", request.environ.get("REQUEST_SCHEME"))
        if request.environ.get("REQUEST_SCHEME") != "https":
            return "reject", "HTTPS scheme required"
            
        u = self.get_user(username)
        #print("_auth_x509: u:", username, u)
        result, reason, expiration = u.authenticate("x509", None, request.environ)
        
        if result:
            return "ok", dict(username=username)
        else:
            return "reject", None
        
    def authenticate(self, method, username, request, redirect):
        #print("AuthCore.authenticate: method:", method)
        try:
            if method == "x509":
                status, extra = self._auth_x509(request, redirect, username)
            elif method == "digest":
                status, extra = self._auth_digest(request.environ, redirect)
            elif method == "ldap":
                status, extra = self._auth_ldap(request, redirect, username)
            elif method == "token":
                status, extra = self._auth_token(request, redirect, username)
            else:
                status, extra = "reject", "Unknown authentication method"
        except:
            traceback.print_exc()
            raise

        return status, extra

    def user_from_request(self, request):
        token, error = self.token_from_request(request)
        #print("AuthenticationCore.user_from_request:", token, token.subject)
        if not token:
            return None, error
        return token.subject, None
            
    def token_from_request(self, request):
        encoded = request.cookies.get("auth_token") or request.headers.get("X-Authentication-Token")
        if not encoded:
            authorization = request.headers.get("Authorization")
            if authorization:
                try:
                    schema, tail = authorization.split(None, 1)
                    if schema.lower() == "bearer":  encoded = tail.strip()
                except:
                    pass
        #print("token_from_request: encoded:", encoded)
        if not encoded: return None, "Token not found"
        return self.verify_token(encoded)



#
# not used
#

from metacat.webapi import MCError
from metacat.util import ObjectSpec
import sys, json, os.path

def catch_mc_errors(method):
    def decorated(*params, **args):
        try:
            return method(*params, **args)
        except MCError as e:
            print(e, file=sys.stderr)
            sys.exit(1)
    return decorated

def load_text(arg):
    text = arg
    if arg:
        file_path = None
        if arg.startswith('@'):             # accept "@<file>" for backward compatibility
            file_path = arg[1:]
        elif os.path.isfile(arg):
            file_path = arg
        if file_path:
            text = open(file_path, "r").read()
        elif arg == "-":
            text = sys.stdin.read()
    return text

def load_json(arg):
    data = None
    text = load_text(arg)
    if text:
        data = json.loads(text)
    return data

def parse_file_spec(spec):
    if isinstance(spec, dict) and "did" in spec:
        ns, n = spec["did"].split(':', 1)
        return {"namespace":ns, "name":n}

def load_file_list(arg):
    text = load_text(arg)
    data = []
    try:
        data = json.loads(text)
    except:
        #print(f"load_file_list: received text: [{text}]")
        for line in text.split("\n"):
            line = line.strip()
            if line:
                for item in line.split(","):
                    item = item.strip()
                    if item:
                        data.append(item)
    #print("load_file_list: data:", data)
    return [ObjectSpec(item).as_dict() for item in data]
        



import sys, getopt, os, json, fnmatch, pprint
from datetime import datetime, timezone
from textwrap import indent
#from urllib.request import urlopen, Request
from metacat.util import to_bytes, to_str, epoch
from metacat.webapi import MetaCatClient, MCError

from metacat.ui.cli import CLI, CLICommand, InvalidArguments


class ListNamedQueriesCommand(CLICommand):

    Opts = "n:j namespace= json"
    Usage = """[options]                     -- list named queries
        -n|--namespace                          - include queries from the namespace only
        -j|--json                               - as JSON
    """

    def __call__(self, command, client, opts, args):
        as_json = "-j" in opts or "--json" in opts
        namespace = opts.get("-n") or opts.get("--namespace")
        queries = client.list_named_queries(namespace=namespace)
        if as_json:
            print(json.dumps(queries, indent=2, sort_keys=True))
        else:
            for query in queries:
                print("Query:      %(namespace)s:%(name)s" % query)
                print("Created by: %(creator)s" % query)
                print("Created on: %s UTC" % (datetime.utcfromtimestamp(query["created_timestamp"]),))
                print("Source:")
                print(indent(query["source"], "  "))
                print()


class ShowNamedQueryCommand(CLICommand):

    Opts = "vj verbose json"
    Usage = """[options] <namespace>:<name>  -- show named query
        -j|--json                               - as JSON
        -v|--verbose                            - verbose outout. Otherwise - source only
    """
    MinArgs = 1
    
    def __call__(self, command, client, opts, args):
        as_json = "-j" in opts or "--json" in opts
        verbose = "-v" in opts or "--verbose" in opts
        
        if ':' not in args[0]:
            raise InvalidArguments()
        namespace, name = args[0].split(':', 1)
        query = client.get_named_query(namespace, name)
        if query is None:
            print("Not found")
            sys.exit(1)
        if as_json:
            print(json.dumps(query, indent=2, sort_keys=True))
        elif not verbose:
            print(query["source"])
        else:
            print("Query:      %(namespace)s:%(name)s" % query)
            print("Created by: %(creator)s" % query)
            print("Created on: %s UTC" % (datetime.utcfromtimestamp(query["created_timestamp"]),))
            print("Source:")
            print(indent(query["source"], "  "))


class CreateNamedQueryCommand(CLICommand):

    Opts = "f:u update file="
    Usage = """                                                -- create named query
        create [options] <namespace>:<name> <MQL query>         - inline query
        create [options] -f|--file <file> <namespace>:<name>    - read query from file
        create [options] <namespace>:<name>                     - read query from stdin
        
        Options:
            -u|--update                -- update if the named query if exists
    """
    MinArgs = 1
    
    def __call__(self, command, client, opts, args):
        if ':' not in args[0]:
            raise InvalidArguments()

        if len(args) > 1:
            query_text = " ".join(args[1:])
        else:
            input_file = opts.get("-f", opts.get("--file", "-"))
            if input_file == "-":
                query_text = sys.stdin.read()
            else:
                query_text = open(input_file, "r").read()

        update = "-u" in opts or "--update" in opts
        namespace, name = args[0].split(':', 1)
        query = client.create_named_query(namespace, name, query_text, update=update)


NamedQueriesCLI = CLI(
    "create",       CreateNamedQueryCommand(),
    "show",         ShowNamedQueryCommand(),
    "list",         ListNamedQueriesCommand()
)
    
 



import sys, getopt, os, json, pprint
from urllib.request import urlopen, Request
from urllib.parse import quote_plus, unquote_plus
from metacat.util import to_bytes, to_str
from metacat.webapi import MetaCatClient, MCServerError, MCWebAPIError, MCError
from metacat.ui.cli import CLICommand, InvalidArguments, InvalidOptions
from metacat.mql import MQLQuery

class CompileCommand(CLICommand):
    
    GNUStyle = False    
    Opts = "q:"
    Usage = """(-q <MQL query file>|"<MQL query>")"""

    def __call__(self, command, client, opts, args):
        query_text = " ".join(args)
        if args:
            query_text = " ".join(args)
        else:
            query_file = opts.get("-q")
            if not query_file:
                raise InvalidArguments("Query must be specified")
            query_text = to_str(open(query_file, "r").read())
        
        print("---- Query text ----\n%s\n" % (query_text,))
        q = MQLQuery.parse(query_text, loader=client)

        print("---- Parsed ----")
        print(q.Parsed.pretty())
        print("")

        print("---- Converted ----")
        print(q.Tree.pretty("    "))
        print("")
        
        q.skip_assembly()
        q.optimize(False)
        print("---- Optimized ----")
        print(q.Optimized.pretty("    "))
        print("")
    
        compiled = q.compile()
        print("---- Compiled ----")
        print(compiled.pretty("    "))

MQLInterpreter = CompileCommand()